/**
 * 
 */
package app;

import java.lang.reflect.InvocationTargetException;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Locale;
import java.util.Random;

import javax.swing.JComboBox;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;

import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.XYPlot;
import org.jfree.data.xy.XYSeries;

import de.congrace.exp4j.Calculable;
import de.congrace.exp4j.ExpressionBuilder;
import de.congrace.exp4j.UnknownFunctionException;
import de.congrace.exp4j.UnparsableExpressionException;

/**
 * Optimizacija funkcija uz ograničenja, Boxov algoritam.
 * 
 * @author Petar
 * 
 */
public class Box implements Algoritam {

	ArrayList<Postavka<?>> postavke;
	final String sep = System.lineSeparator();
	double eps;
	double alfa;
	double donjaGranicaX, gornjaGranicaX, donjaGranicaY, gornjaGranicaY;
	String ekspl_x, ekspl_y, impl_1, impl_2;
	Matrix xp, Xc, Xr;
	Matrix[] tocke;
	Calculable prva, druga, funkcija;
	boolean manje1, manje2, next=false, postoji1=true, postoji2=true;
	int dimenzije = 2;
	double cekaj;
	DecimalFormat df;
	State stanje;
	long msec;
	int PAUSE=2,NEXT=3;
	
	/**
	 * Konstruktor
	 * 
	 * @throws UnknownFunctionException
	 * @throws UnparsableExpressionException
	 */
	public Box() throws UnknownFunctionException, UnparsableExpressionException {

		NumberFormat nf = NumberFormat.getNumberInstance(Locale.ENGLISH);
		df = (DecimalFormat) nf;
		df.setMinimumFractionDigits(2);
		df.setMaximumFractionDigits(2);
		df.setMinimumIntegerDigits(1);
		df.setMaximumIntegerDigits(3);
		df.setGroupingSize(20);

		this.alfa = 1.3d;
		this.eps = 0.001d;
		this.xp = new Matrix(1.0d, 1.0d);
		tocke = new Matrix[2 * dimenzije];

		for (int i = 0; i < 2 * dimenzije; i++)
			tocke[i] = xp.dajKopiju();

		this.impl_1 = "x-y <= 0";
		this.impl_2 = "x-2 <= 0";
		this.ekspl_x = "-100,100";
		this.ekspl_y = "-100,100";

		postavke = new ArrayList<Postavka<?>>();
		postavke.add(new Postavka<Double>("koeficijent refleksije", alfa));
		postavke.add(new Postavka<Double>("epsilon", eps));
		postavke.add(new Postavka<Matrix>("Xp", xp));
		postavke.add(new Postavka<String>("implicitno ograničenje", impl_1));
		postavke.add(new Postavka<String>("implicitno ograničenje", impl_2));
		postavke.add(new Postavka<String>("eksplicitno ograničenje x", ekspl_x));
		postavke.add(new Postavka<String>("eksplicitno ograničenje y", ekspl_y));

		rješiPostavke();

	}

	/**
	 * Priprema postavke i varijable razreda
	 */
	private void rješiPostavke() {
		this.alfa = (Double) postavke.get(0).Vrijednost;
		this.eps = (Double) postavke.get(1).Vrijednost;
		this.xp = (Matrix) postavke.get(2).Vrijednost;
		this.impl_1 = (String) postavke.get(3).Vrijednost;
		this.impl_2 = (String) postavke.get(4).Vrijednost;
		this.ekspl_x = (String) postavke.get(5).Vrijednost;
		this.ekspl_y = (String) postavke.get(6).Vrijednost;

		String[] split = ekspl_x.split(",");
		String[] split2 = ekspl_y.split(",");

		donjaGranicaX = Double.parseDouble(split[0]);
		gornjaGranicaX = Double.parseDouble(split[1]);
		donjaGranicaY = Double.parseDouble(split2[0]);
		gornjaGranicaY = Double.parseDouble(split2[1]);
		
		try {
			if (this.impl_1.contains("<=")) {
				manje1 = true;
				prva = new ExpressionBuilder(impl_1.split("<=")[0])
						.withVariableNames("x", "y").build();

			} else if (this.impl_1.contains(">=")){
				manje1 = false;
				prva = new ExpressionBuilder(impl_1.split(">=")[0])
						.withVariableNames("x", "y").build();
			}
			else if (this.impl_1.equalsIgnoreCase("")) postoji1=false;

			if (this.impl_2.contains("<=")) {
				manje2 = true;
				druga = new ExpressionBuilder(impl_2.split("<=")[0])
						.withVariableNames("x", "y").build();

			} else if (this.impl_2.contains(">=")){
				manje2 = false;
				druga = new ExpressionBuilder(impl_2.split(">=")[0])
						.withVariableNames("x", "y").build();
			}
			else if (this.impl_2.equalsIgnoreCase("")) postoji2=false;

		} catch (UnknownFunctionException | UnparsableExpressionException e) {
			throw new IllegalArgumentException("Greška pri unosu parametara!" + sep
					+ e.getMessage());
		}
		
		

	}

	/**
	 * Provjera implicitnih ograničenja
	 * 
	 * @param tocka
	 *            točka koja se provjerava
	 * @return true/false
	 */
	private boolean provjeriImplicitna(Matrix tocka) {
		double x = tocka.dohvati(0, 0);
		double y = tocka.dohvati(0, 1);
		prva.setVariable("x", x);
		prva.setVariable("y", y);
		druga.setVariable("x", x);
		druga.setVariable("y", y);
		double result1 = prva.calculate();
		double result2 = druga.calculate();
		
		if (postoji1){
		if (manje1 && result1 > 0)
			return false;
		if (!manje1 && result1 < 0)
			return false;
		}
		if (postoji2){
		if (manje2 && result2 > 0)
			return false;
		if (!manje2 && result2 < 0)
			return false;
		}
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see app.Algoritam#dajPostavke()
	 */
	@Override
	public ArrayList<Postavka<?>> dajPostavke() {
		return postavke;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see app.Algoritam#spremiPostavke()
	 */
	@Override
	public void spremiPostavke(ArrayList<String> lista_gui) throws InvocationTargetException {

		try {

			this.alfa = Double.parseDouble(lista_gui.get(0));
			this.eps = Double.parseDouble(lista_gui.get(1));
			String[] split = lista_gui.get(2).split(",");
			this.xp = new Matrix(Double.parseDouble(split[0].trim()),
					Double.parseDouble(split[1].trim()));
			this.impl_1 = lista_gui.get(3);
			this.impl_2 = lista_gui.get(4);
			this.ekspl_x = lista_gui.get(5);
			this.ekspl_y = lista_gui.get(6);

			postavke.clear();
			postavke.add(new Postavka<Double>("koeficijent refleksije", alfa));
			postavke.add(new Postavka<Double>("epsilon", eps));
			postavke.add(new Postavka<Matrix>("Xp", xp));
			postavke.add(new Postavka<String>("implicitno ograničenje", impl_1));
			postavke.add(new Postavka<String>("implicitno ograničenje", impl_2));
			postavke.add(new Postavka<String>("eksplicitno ograničenje x",
					ekspl_x));
			postavke.add(new Postavka<String>("eksplicitno ograničenje y",
					ekspl_y));

			rješiPostavke();

		} catch (IllegalArgumentException e) {
			throw new IllegalArgumentException("Greška pri unosu parametara!" + sep
					+ e.getMessage());
		}

		if (!(provjeriEks(xp) && provjeriImplicitna(xp))) {

			throw new IllegalArgumentException("Početna točka ne zadovoljava ograničenja!");

		}

	}

	@Override
	public String getName() {
		return "Box";
	}

	@Override
	public void postaviFunkciju(String tekst) throws IllegalArgumentException {
		try {
			funkcija = new ExpressionBuilder(tekst).withVariableNames("x", "y")
					.build();
		} catch (Exception e) {
			throw new IllegalArgumentException("Greška pri unosu funkcije!" + sep
					+ e.getMessage());
		}

	}

	@Override
	public void kreni(final Prikaz p, final JTextArea ispis,
			final JComboBox<String> delay, State status)
			throws InvocationTargetException, InterruptedException {
		
		
		this.stanje=status;
		Random rand = new Random();
		Matrix X0 = xp.dajKopiju();
		Matrix pom = new Matrix(1, dimenzije);

		Xc = X0.dajKopiju();
		tocke[0] = X0;
		// stvaranje 2n tocaka
		for (int t = 1; t < 2 * dimenzije; t++) {

			tocke[t].polje[0][0] = donjaGranicaX + rand.nextDouble()
					* (gornjaGranicaX - donjaGranicaX);
			tocke[t].polje[0][1] = donjaGranicaY + rand.nextDouble()
					* (gornjaGranicaY - donjaGranicaY);
			while (true) {
				if (provjeriImplicitna(tocke[t]))
					break;
				tocke[t] = tocke[t].zbrojiSa(Xc);
				tocke[t] = tocke[t].mnoziSaSkalarom(0.5);

			}

			// racunanje Xc
			Matrix suma = pom.dajKopiju();
			for (int k = 0; k <= t; k++) {
				suma = suma.zbrojiSa(tocke[k]);
			}
			suma = suma.mnoziSaSkalarom(1.0 / (t + 1));
			Xc = suma;
		}

		

		SwingUtilities.invokeAndWait(new Runnable() {

			@Override
			public void run() {
				cekaj = Double.parseDouble((String) delay.getSelectedItem());
				ispis.append("Dobrodošli, pozdravlja Vas Boxov algoritam!!"
						+ sep);
				ispis.append("Lets go!" + sep);
				ispis.append("========================================" + sep);
				ispis.append("Točke:" + sep);
				ispis.append("--------------------------------------------------------------------------"
						+ sep);
				for (Matrix t : tocke)
					ispis.append(print(t) + "\t");
				ispis.append(sep);

				// crtanje

				p.dataset.removeAllSeries();
				p.podaci.clear();

				p.podaci.add(new XYSeries("tocke"));
				p.podaci.add(new XYSeries("Xc"));
				p.podaci.add(new XYSeries("Xr"));
				p.podaci.add(new XYSeries("1. implicitno ograničenje"));
				p.podaci.add(new XYSeries("2. implicitno ograničenje"));

				for (Matrix t : tocke)
					p.podaci.get(0).add(t.polje[0][0], t.polje[0][1]);

				p.dataset.addSeries(p.podaci.get(0));
				p.dataset.addSeries(p.podaci.get(1));
				p.dataset.addSeries(p.podaci.get(2));

				p.jf.setTitle("Box");

				final XYPlot plot = p.jf.getXYPlot();
				plot.setDomainZeroBaselineVisible(true);
				plot.setRangeZeroBaselineVisible(true);
				ValueAxis axis = plot.getDomainAxis();
				axis.setAutoRange(true);
				axis.setAutoRangeMinimumSize(20.0);
				axis = plot.getRangeAxis();
				axis.setAutoRange(true);
				axis.setAutoRangeMinimumSize(20.0);
				
				
			}
		});

		while (true) {
			
			justDoIt();

			// nadji h i h2
			double max = Double.MIN_VALUE;
			int h = 0;
			int h2 = 0;
			for (int i = 0; i < 2 * dimenzije; i++) {
				if (racunaj(tocke[i]) > max) {
					h = i;
					max = racunaj(tocke[i]);
				}

			}
			max = Double.MIN_VALUE;
			for (int i = 0; i < 2 * dimenzije; i++) {
				if (racunaj(tocke[i]) > max && i != h) {
					h2 = i;
					max = racunaj(tocke[i]);
				}
			}

			Matrix suma = pom.dajKopiju();
			for (int k = 0; k < 2 * dimenzije; k++) {
				if (k == h)
					continue;
				suma = suma.zbrojiSa(tocke[k]);
			}
			suma = suma.mnoziSaSkalarom(1.0 / (2 * dimenzije - 1));
			Xc = suma;

			final int kvbrc = h;

			SwingUtilities.invokeAndWait(new Runnable() {

				@Override
				public void run() {
					cekaj = Double.parseDouble((String) delay.getSelectedItem());
					ispis.append("Najgora točka je " + print(tocke[kvbrc])
							+ sep);
					ispis.append("Xc: " + print(Xc) + sep);

					p.podaci.get(1).add(Xc.polje[0][0], Xc.polje[0][1]);
				}
			});
			
			
			justDoIt();

			// traži Xr
			Xr = pom.dajKopiju();
			Xr = (tocke[h].mnoziSaSkalarom(alfa)).oduzmiOd(Xc
					.mnoziSaSkalarom(1.0 + alfa));

			SwingUtilities.invokeAndWait(new Runnable() {

				@Override
				public void run() {
					cekaj = Double.parseDouble((String) delay.getSelectedItem());
					ispis.append("Xr: " + print(Xr) + sep);

					p.podaci.get(2).add(Xr.polje[0][0], Xr.polje[0][1]);

				}
			});
			justDoIt();

			if (Xr.polje[0][0] < donjaGranicaX)
				Xr.polje[0][0] = donjaGranicaX;
			else if (Xr.polje[0][0] > gornjaGranicaX)
				Xr.polje[0][0] = gornjaGranicaX;

			if (Xr.polje[0][1] < donjaGranicaY)
				Xr.polje[0][1] = donjaGranicaY;
			else if (Xr.polje[0][1] > gornjaGranicaY)
				Xr.polje[0][1] = gornjaGranicaY;

			SwingUtilities.invokeAndWait(new Runnable() {

				@Override
				public void run() {
					cekaj = Double.parseDouble((String) delay.getSelectedItem());
					ispis.append("Provjera eksplicitnih ograničenja, Xr: "
							+ print(Xr) + sep);

					p.podaci.get(2).clear();
					p.podaci.get(2).add(Xr.polje[0][0], Xr.polje[0][1]);

				}
			});
			justDoIt();

			while (true) {
				if (provjeriImplicitna(Xr))
					break;
				Xr = (Xr.zbrojiSa(Xc)).mnoziSaSkalarom(0.5);
			}

			if (racunaj(Xr) > racunaj(tocke[h2]))
				Xr = (Xr.zbrojiSa(Xc)).mnoziSaSkalarom(0.5);

			SwingUtilities.invokeAndWait(new Runnable() {

				@Override
				public void run() {
					cekaj = Double.parseDouble((String) delay.getSelectedItem());
					ispis.append("Provjera implicitnih ograničenja, Xr: "
							+ print(Xr) + sep);

					p.podaci.get(2).clear();
					p.podaci.get(2).add(Xr.polje[0][0], Xr.polje[0][1]);

				}
			});
			
			justDoIt();
			
			
			final Matrix trc = tocke[h].dajKopiju();
			tocke[h] = Xr;
			SwingUtilities.invokeAndWait(new Runnable() {

				@Override
				public void run() {
					cekaj = Double.parseDouble((String) delay.getSelectedItem());
					ispis.append("Zamjenjujem točku " + print(trc)
							+ " sa točkom " + print(Xr) + sep);
					ispis.append("--------------------------------------------------------------------------"
							+ sep);
					for (Matrix t : tocke)
						ispis.append(print(t) + "\t");
					ispis.append(sep);

					p.podaci.get(0).clear();
					p.podaci.get(1).clear();
					p.podaci.get(2).clear();

					for (Matrix t : tocke)
						p.podaci.get(0).add(t.polje[0][0], t.polje[0][1]);

				}
			});
			
			// provjeri kriterij zaustavljanja
			double uk = 0;
			for (int i = 0; i < 2 * dimenzije; i++) {
				uk += Math.pow(racunaj(tocke[i]) - racunaj(Xc), 2);
			}
			uk = uk / (2.0 * dimenzije);
			uk = Math.pow(uk, 0.5);
			if (uk <= eps) {
				SwingUtilities.invokeAndWait(new Runnable() {

					@Override
					public void run() {
						cekaj = Double.parseDouble((String) delay
								.getSelectedItem());
						ispis.append("--------------------------------------------------------------------------"
								+ sep);
						ispis.append("Gotovo!!!" + sep);

					}
				});
				break;
			}
		}

	}

	private void justDoIt() {
		obradiStanje();
		if (!next){
		msec = (long) (cekaj * 1000.0d);
		try {
			Thread.sleep(msec);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		}
		
	}

	private  void obradiStanje() {
			
			synchronized (stanje) {
				int result = stanje.getStanje();
				if (next) {
					result=PAUSE;
					next=false;
				}
				if (result==PAUSE){
					try {
						stanje.wait();
					} catch (InterruptedException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
				if (stanje.getStanje()==NEXT){
					next=true;
				}
				
			}
		
			
		
		
	}

	

	/**
	 * Računa vrijednost funkcije u točki x
	 * 
	 * @param tocka
	 *            točka x
	 * @return vrijednost funkcije u točki x
	 */
	private double racunaj(Matrix tocka) {
		funkcija.setVariable("x", tocka.polje[0][0]);
		funkcija.setVariable("y", tocka.polje[0][1]);
		return funkcija.calculate();
	}

	/**
	 * Provjera eksplicitnih ograničenja
	 * 
	 * @param tocka
	 *            točka koja se provjerava
	 * @return true/false
	 */
	private boolean provjeriEks(Matrix tocka) {
		double x = tocka.dohvati(0, 0);
		double y = tocka.dohvati(0, 1);
		if (x < donjaGranicaX || x > gornjaGranicaX || y < donjaGranicaY
				|| y > gornjaGranicaY)
			return false;
		return true;
	}

	/**
	 * Prilagođen ispis matrica za GUI
	 * 
	 * @param x
	 * @return
	 */
	protected String print(Matrix x) {
		return df.format(x.polje[0][0]) + ", " + df.format(x.polje[0][1]);
	}

}
