package app;

import java.lang.reflect.InvocationTargetException;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Locale;

import javax.swing.JComboBox;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;

import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.XYPlot;
import org.jfree.data.xy.XYSeries;

import de.congrace.exp4j.Calculable;
import de.congrace.exp4j.ExpressionBuilder;
import de.congrace.exp4j.UnknownFunctionException;
import de.congrace.exp4j.UnparsableExpressionException;

/**
 * Pronalaženje optimuma - hooke-jeeves metodom!
 * @author Petar
 *
 */
public class HookeJeeves implements Algoritam {

	Matrix dx, e, Xp, Xn, Xb;
	ArrayList<Postavka<?>> postavke;
	Calculable calc;
	DecimalFormat df;
	double cekaj;
	final String sep = System.lineSeparator();
	boolean dxmanji;
	State stanje;
	boolean next;
	int PAUSE=2,NEXT=3;
	long msec;

	/**
	 * Konstruktor
	 */
	public HookeJeeves() {
		this.dx = new Matrix(0.5, 0.5);
		this.e = new Matrix(0.05, 0.05);
		this.Xp = new Matrix(1.0d, 1.0d);

		postavke = new ArrayList<Postavka<?>>();

		postavke.add(new Postavka<Matrix>("dx", dx));
		postavke.add(new Postavka<Matrix>("epsilon", e));
		postavke.add(new Postavka<Matrix>("Xp", Xp));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see app.Algoritam#dajPostavke()
	 */
	@Override
	public ArrayList<Postavka<?>> dajPostavke() {
		return postavke;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see app.Algoritam#spremiPostavke()
	 */
	@Override
	public void spremiPostavke(ArrayList<String> lista_gui) throws IllegalArgumentException {

		try {

			for (int i = 0; i < lista_gui.size(); i++) {
				String[] split = lista_gui.get(i).split(",");
				Matrix s = (Matrix) postavke.get(i).Vrijednost;
				s.polje[0][0] = Double.parseDouble(split[0].trim());
				s.polje[0][1] = Double.parseDouble(split[1].trim());
			}
			dx = (Matrix) postavke.get(0).Vrijednost;
			e = (Matrix) postavke.get(1).Vrijednost;
			Xp = (Matrix) postavke.get(2).Vrijednost;
		} catch (IllegalArgumentException e) {
			throw new IllegalArgumentException("Greška pri unosu parametara!" + sep
					+ e.getMessage());
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see app.Algoritam#kreni(app.Prikaz)
	 */
	@Override
	public void kreni(final Prikaz p, final JTextArea ispis,
			final JComboBox<String> delay, State status)
			throws InvocationTargetException, InterruptedException {

		Xb = Xp.dajKopiju();
		Xn = Xp.dajKopiju();
		this.stanje=status;
		
		NumberFormat nf = NumberFormat.getNumberInstance(Locale.ENGLISH);
		df = (DecimalFormat) nf;
		df.setMinimumFractionDigits(3);
		df.setMaximumFractionDigits(3);
		df.setMinimumIntegerDigits(1);
		df.setMaximumIntegerDigits(3);
		df.setGroupingSize(20);

		SwingUtilities.invokeAndWait(new Runnable() {

			@Override
			public void run() {

				p.dataset.removeAllSeries();
				p.podaci.clear();

				ispis.append("Dobrodošli, pozdravlja Vas Hooke-Jeeves algoritam!!"
						+ sep);
				ispis.append("Lets go!" + sep);

				ispis.append("========================================" + sep);
				ispis.append("Vrijednost dx: " + print(dx) + sep);
				ispis.append("            Xb                  Xp                 Xn"
						+ sep);
				ispis.append("-------------------------------------------------------"
						+ sep);
				ispis.append("    " + print(Xb) + "    " + print(Xp) + "    ");

				// crtanje..

				p.podaci.add(new XYSeries("Xb"));
				p.podaci.add(new XYSeries("Xp"));
				p.podaci.add(new XYSeries("Xn"));

				p.podaci.get(1).add(Xp.polje[0][0], Xp.polje[0][1]);
				p.podaci.get(0).add(Xb.polje[0][0], Xb.polje[0][1]);

				p.dataset.addSeries(p.podaci.get(0));
				p.dataset.addSeries(p.podaci.get(1));
				p.dataset.addSeries(p.podaci.get(2));

				p.jf.setTitle("Hooke-Jeeves");
				final XYPlot plot = p.jf.getXYPlot();
				plot.setDomainZeroBaselineVisible(true);
				plot.setRangeZeroBaselineVisible(true);
				ValueAxis axis = plot.getDomainAxis();
				axis.setAutoRange(true);
				axis.setAutoRangeMinimumSize(10.0);
				axis = plot.getRangeAxis();
				axis.setAutoRange(true);
				axis.setAutoRangeMinimumSize(10.0);

				cekaj = Double.parseDouble((String) delay.getSelectedItem());
			}
		});

		while (true) {

			justDoIt();
			nadjiNoviXn();

			SwingUtilities.invokeAndWait(new Runnable() {

				@Override
				public void run() {
					if (Xn.isti(Xb))
						p.podaci.get(0).clear();
					if (Xn.isti(Xp))
						p.podaci.get(1).clear();
					p.podaci.get(2).add(Xn.polje[0][0], Xn.polje[0][1]);
					cekaj = Double.parseDouble((String) delay.getSelectedItem());
					ispis.append(print(Xn) + sep);
				}
			});

			justDoIt();
			dxmanji = false;

			if (racunaj(Xn) < racunaj(Xb)) {
				Xp = Xb.oduzmiOd(Xn.mnoziSaSkalarom(2.0));
				Xb = Xn;
			} else {
				dx = dx.mnoziSaSkalarom(0.5);
				Xp = Xb;
				dxmanji = true;
			}
			SwingUtilities.invokeAndWait(new Runnable() {

				@Override
				public void run() {

					if (dxmanji) {
						ispis.append("-------------------------------------------------------"
								+ sep);
						ispis.append("Nova vrijednost dx: " + print(dx) + sep);
						ispis.append("-------------------------------------------------------"
								+ sep);
					}

					ispis.append("    " + print(Xb) + "    " + print(Xp)
							+ "    ");
					p.podaci.get(0).clear();
					p.podaci.get(1).clear();
					p.podaci.get(2).clear();

					p.podaci.get(0).add(Xb.polje[0][0], Xb.polje[0][1]);
					p.podaci.get(1).add(Xp.polje[0][0], Xp.polje[0][1]);

					cekaj = Double.parseDouble((String) delay.getSelectedItem());
				}
			});


			if (provjeri()) {
				SwingUtilities.invokeAndWait(new Runnable() {

					@Override
					public void run() {
						ispis.append(sep);
						ispis.append("-------------------------------------------------------"
								+ sep);
						ispis.append("Rezultat: " + print(Xb) + sep);

					}
				});
				break;
			}
		}

	}

	/**
	 * Prilagođen ispis matrica za GUI
	 * 
	 * @param x
	 * @return
	 */
	protected String print(Matrix x) {
		return df.format(x.polje[0][0]) + ", " + df.format(x.polje[0][1]);
	}

	/**
	 * Računa vrijednost funkcije u točki x
	 * 
	 * @param tocka
	 *            točka x
	 * @return vrijednost funkcije u točki x
	 */
	private double racunaj(Matrix x) {
		calc.setVariable("x", x.polje[0][0]);
		calc.setVariable("y", x.polje[0][1]);
		return calc.calculate();
	}

	/**
	 * Provjera da li je algoritam gotov
	 * 
	 * @return true ako treba zaustaviti algoritam
	 */
	private boolean provjeri() {
		for (int i = 0; i < dx.polje[0].length; i++) {
			if (dx.polje[0][i] > e.polje[0][i])
				return false;
		}
		return true;
	}

	/**
	 * Nalazi novi Xn u okolici Xp
	 */
	private void nadjiNoviXn() {

		Matrix x = Xp.dajKopiju();
		for (int i = 0; i < Xn.polje[0].length; i++) {

			double pe = racunaj(x);
			x.polje[0][i] += dx.polje[0][i];
			double en = racunaj(x);
			if (en > pe) {
				x.polje[0][i] -= 2.0 * dx.polje[0][i];
				en = racunaj(x);
				if (en > pe)
					x.polje[0][i] += dx.polje[0][i];
			}
		}
		Xn = x;

	}
	
	
	private void justDoIt() {
		obradiStanje();
		if (!next){
		msec = (long) (cekaj * 1000.0d);
		try {
			Thread.sleep(msec);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		}
		
	}
	
	private  void obradiStanje() {
		
		synchronized (stanje) {
			int result = stanje.getStanje();
			if (next) {
				result=PAUSE;
				next=false;
			}
			if (result==PAUSE){
				try {
					stanje.wait();
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			if (stanje.getStanje()==NEXT){
				next=true;
			}
			
		}
	
		
	
	
}
	
	

	@Override
	public String getName() {
		return "Hooke Jeeves";
	}

	public void postaviFunkciju(String tekst) throws IllegalArgumentException {
		try {
			calc = new ExpressionBuilder(tekst).withVariableNames("x", "y")
					.build();
		} catch (UnknownFunctionException | UnparsableExpressionException e) {
			throw new IllegalArgumentException("Greška pri unosu funkcije!" + sep
					+ e.getMessage());
		}
	}

}
